@@ -10,16 +10,17 @@ use autodie;
 use MooX qw/late/;
 use URI;
 use AnyEvent::HTTP qw/http_head http_get/;
+use AnyEvent::FTP::Client;
 use Getopt::Long qw/GetOptionsFromArray/;
 use File::Basename qw(basename);
 use Fcntl qw( SEEK_SET );
 use List::UtilsBy qw(max_by);
-use JSON qw(decode_json encode_json);
+use JSON::MaybeXS qw(decode_json encode_json);
 
 use App::ManiacDownloader::_SegmentTask;
 use App::ManiacDownloader::_BytesDownloaded;
 
-our $VERSION = '0.0.10';
+our $VERSION = '0.0.11';
 
 my $DEFAULT_NUM_CONNECTIONS = 4;
 my $NUM_CONN_BYTES_THRESHOLD = 4_096 * 2;
@@ -27,7 +28,9 @@ my $NUM_CONN_BYTES_THRESHOLD = 4_096 * 2;
 has '_finished_condvar' => (is => 'rw');
 has '_ranges' => (isa => 'ArrayRef', is => 'rw');
 has '_url' => (is => 'rw');
+has '_url_path' => (isa => 'Str', is => 'rw');
 has '_url_basename' => (isa => 'Str', is => 'rw');
+has '_is_ftp' => (isa => 'Bool', is => 'rw');
 has '_remaining_connections' => (isa => 'Int', is => 'rw');
 has '_stats_timer' => (is => 'rw');
 has '_last_timer_time' => (is => 'rw', isa => 'Num');
@@ -69,10 +72,19 @@ sub _start_connection
 
     sysseek( $r->_fh, $r->_start, SEEK_SET );
 
+    my $is_ftp = $self->_is_ftp;
+
     # We do these to make sure the cancellation guard does not get
     # preserved because it's in the context of the closures.
     my $on_body = sub {
-        my ($data, $hdr) = @_;
+        my ($active_seq, $data, $hdr) = @_;
+
+        # Stale or wrong connection - probably AnyEvent::FTP::Client after a
+        # quit.
+        if ((! $r->_is_right_active_seq($active_seq)) or (! $r->is_active))
+        {
+            return;
+        }
 
         my $ret = $r->_write_data(\$data);
 
@@ -81,6 +93,10 @@ sub _start_connection
         my $cont = $ret->{should_continue};
         if (! $cont)
         {
+            if ($is_ftp)
+            {
+                $r->_guard->quit;
+            }
             my $largest_r = max_by { $r->_num_remaining } @{$self->_ranges};
             if ($largest_r->_num_remaining < $NUM_CONN_BYTES_THRESHOLD)
             {
@@ -106,16 +122,45 @@ sub _start_connection
 
     my $final_cb = sub { return ; };
 
-    my $guard = http_get $self->_url,
-    headers => { 'Range'
-        => sprintf("bytes=%d-%d", $r->_start, $r->_end-1)
-    },
-    on_body => $on_body,
-    $final_cb;
+    my $url = $self->_url;
+    {
+        my $active_seq = $r->_get_next_active_seq;
 
-    $r->_guard($guard);
+        my $seq_on_body = sub { return $on_body->($active_seq, @_); };
 
-    $guard = '';
+        if ($is_ftp)
+        {
+            my $ftp = AnyEvent::FTP::Client->new( passive => 1 );
+            $r->_guard($ftp);
+            $ftp->connect($url->host, $url->port)->cb(sub {
+                    $ftp->login($url->user, $url->password)->cb(sub {
+                            $ftp->type('I')->cb(sub {
+                                    $ftp->retr(
+                                        $self->_url_path,
+                                        $seq_on_body,
+                                        restart => $r->_start,
+                                    );
+                                }
+                            );
+                        }
+                    );
+                }
+            );
+        }
+        else
+        {
+            my $guard = http_get $url,
+            headers => { 'Range'
+                => sprintf("bytes=%d-%d", $r->_start, $r->_end-1)
+            },
+            on_body => $seq_on_body,
+            $final_cb;
+
+            $r->_guard($guard);
+
+            $guard = '';
+        }
+    }
 
     return;
 }
@@ -197,6 +242,24 @@ sub _open_fh_for_read_write_without_clobbering
     return $fh;
 }
 
+sub _with_len_and_num_connections
+{
+    my ($self, $len, $num_connections) = @_;
+
+    if (!defined($len)) {
+        die "Cannot find a content-length header.";
+    }
+
+    $self->_len($len);
+    $self->_remaining_connections($num_connections);
+
+    return $self->_init_from_len(
+        {
+            num_connections => $num_connections,
+        }
+    );
+}
+
 sub _init_from_len
 {
     my ($self, $args) = @_;
@@ -298,13 +361,17 @@ sub run
         or die "No url given.";
 
     my $url = URI->new($url_s);
-
     $self->_url($url);
-    my $url_path = $url->path();
-    my $url_basename = basename($url_path);
 
+    my $url_path = $url->path();
+    $self->_url_path($url_path);
+
+    my $url_basename = basename($url_path);
     $self->_url_basename($url_basename);
 
+    my $is_ftp = ($url->scheme eq 'ftp');
+    $self->_is_ftp($is_ftp);
+
     if (-e $self->_url_basename)
     {
         print STDERR "File appears to have already been downloaded. Quitting.\n";
@@ -331,23 +398,34 @@ sub run
     }
     else
     {
-        http_head $url, sub {
-            my (undef, $headers) = @_;
-            my $len = $headers->{'content-length'};
+        if ($is_ftp)
+        {
+            my $ftp = AnyEvent::FTP::Client->new( passive => 1 );
+            $ftp->connect($url->host, $url->port)->recv;
+            $ftp->login($url->user, $url->password)->recv;
+            $ftp->type('I')->recv;
+            $ftp->size( $self->_url_path)->cb(
+                sub {
+                    my $len = shift->recv;
 
-            if (!defined($len)) {
-                die "Cannot find a content-length header.";
-            }
+                    $ftp->quit;
+                    undef($ftp);
 
-            $self->_len($len);
-            $self->_remaining_connections($num_connections);
-
-            return $self->_init_from_len(
-                {
-                    num_connections => $num_connections,
+                    return $self->_with_len_and_num_connections(
+                        $len, $num_connections
+                    );
                 }
             );
-        };
+        }
+        else
+        {
+            http_head $url, sub {
+                my (undef, $headers) = @_;
+                my $len = $headers->{'content-length'};
+
+                return $self->_with_len_and_num_connections($len, $num_connections);
+            };
+        }
     }
 
     my $signal_handler = sub { $self->_abort_signal_handler(); };
@@ -379,7 +457,7 @@ App::ManiacDownloader - a maniac download accelerator.
 
 =head1 VERSION
 
-version 0.0.10
+version 0.0.11
 
 =head1 SYNOPSIS
 
